@@ -1,5 +1,5 @@
 package Bio::Das;
-# $Id: Das.pm,v 1.23 2003/12/29 23:20:32 lstein Exp $
+# $Id: Das.pm,v 1.28 2004/01/12 23:29:00 lstein Exp $
 
 # prototype parallel-fetching Das
 
@@ -24,8 +24,8 @@ use IO::Select;
 use vars '$VERSION';
 use vars '@ISA';
 @ISA     = 'Bio::Root::Root';
-$VERSION = 0.93;
-*fetch_feature_by_name = \&get_feature_by_name;
+$VERSION = 0.94;
+*feature2segment = *fetch_feature_by_name = \&get_feature_by_name;
 my @COLORS = qw(cyan blue red yellow green wheat turquoise orange);
 
 sub new {
@@ -38,7 +38,7 @@ sub new {
   if (@p >= 1 && $p[0] =~ /^http/) {
     ($url,$dsn,$aggregators) = @p;
   } elsif ($p[0] =~ /^-/) {  # named arguments
-    ($url,$dsn,$aggregators,$timeout,$auth_callback) = rearrange(['source',
+    ($url,$dsn,$aggregators,$timeout,$auth_callback) = rearrange([['source','server'],
 								  'dsn',
 								  ['aggregators','aggregator'],
 								  'timeout',
@@ -118,7 +118,11 @@ sub clear_aggregators { shift->{aggregators} = [] }
 sub default_dsn {
   my $self = shift;
   my $d    = $self->{default_dsn};
-  $self->{default_dsn} = shift if @_;
+  if (@_) {
+    my $new_dsn = shift;
+    $self->{default_dsn} = UNIVERSAL::isa($new_dsn,'Bio::Das::DSN') ?
+                           $new_dsn->id : $new_dsn;
+  }
   $d;
 }
 
@@ -144,6 +148,13 @@ sub auth_callback{
   $d;
 }
 
+sub no_rfc_warning {
+  my $self = shift;
+  my $d   = $self->{no_rfc_warning};
+  $self->{no_rfc_warning} = shift if @_;
+  $d;
+}
+
 sub proxy {
   my $self = shift;
   my $d    = $self->{proxy};
@@ -165,6 +176,17 @@ sub debug {
   $d;
 }
 
+sub make_fetcher {
+  my $self    = shift;
+  my $request = shift;
+  return Bio::Das::HTTP::Fetch->new(
+				    -request   => $request,
+				    -headers   => {'Accept-encoding' => 'gzip'},
+				    -proxy     => $self->proxy || '',
+				    -norfcwarn => $self->no_rfc_warning,
+				   );
+}
+
 # call with list of base names
 # will return a list of DSN objects
 sub dsn {
@@ -175,8 +197,15 @@ sub dsn {
 
 sub _dsn {
   my $self = shift;
-  my @requests = $_[0]=~/^-/ ? Bio::Das::Request::Dsn->new(@_)
-                             : map { Bio::Das::Request::Dsn->new($_) } @_;
+  my @dsns;
+  if ($_[0] =~ /^-/) {
+    my($dsn) = rearrange([['dsn','dsns']],@_);
+    @dsns    = ref($dsn) eq 'ARRAY' ? @$dsn : ($dsn);
+  }
+  else {
+    @dsns = @_;
+  }
+  my @requests = map { Bio::Das::Request::Dsn->new($_) } @dsns;
   $self->run_requests(\@requests);
 }
 
@@ -288,16 +317,24 @@ sub segment {
 # 0.18 API - fetch by feature name - returns a set of Bio::Das::Segment objects
 sub get_feature_by_name {
   my $self = shift;
-  my ($name, $class, $ref, $base_start, $stop) 
-       = $self->_rearrange([qw(NAME CLASS REF START END)],@_);
-  my $dsn = $self->default_url;
-  my $request = Bio::Das::Request::Feature2Segments->new(-class   => $class,
-							 -dsn     => $dsn,
-							 -feature => $name,
-							 -das     => $self,
-							);
-  $self->run_requests([$request]);
-  return $request->results;
+  my ($class, $name, $dsn);
+  if (@_ == 1) {
+    $name = shift;
+  } else {
+    ($class, $name, $dsn)
+      = $self->_rearrange([qw(CLASS NAME DSN)],@_);
+  }
+  $dsn ||= $self->default_url;
+  croak "must provide -dsn argument" unless $dsn;
+  my @dsn = ref $dsn && ref $dsn eq 'ARRAY' ? @$dsn : $dsn;
+
+  my @requests = map { Bio::Das::Request::Feature2Segments->new(-class   => $class,
+								-dsn     => $_,
+								-feature => $name,
+								-das     => $self,
+							       )
+  } @dsn;
+  $self->run_requests(\@requests);
 }
 
 # gbrowse compatibility
@@ -330,7 +367,7 @@ sub features {
     unshift @aggregators,$a if $a->disaggregate($typearray,$typehandler);
   }
 
-  my @types = map {$_->[0]} @$typearray;
+  my @types = map {defined $_->[1] ? "$_->[0]:$_->[1]" : $_->[0]} @$typearray;
   my @request;
   for my $dsn (@dsn) {
     push @request,Bio::Das::Request::Features->new(
@@ -348,7 +385,6 @@ sub features {
   $self->aggregate(\@aggregators,
 		   $results[0]->can('results') ? \@results : [\@results],
 		   $typehandler) if @aggregators && @results;
-
   return Bio::Das::FeatureIterator->new(\@results) if $iterator;
   return wantarray ? @results : $results[0];
 }
@@ -390,22 +426,16 @@ sub remove_pending {
 sub run_requests {
   my $self     = shift;
   my $requests = shift;
-  my $auth_callback = $self->auth_callback();
 
   for my $request (@$requests) {
-    my $fetcher = Bio::Das::HTTP::Fetch->new(
-                            -request => $request,
-					        -headers => {'Accept-encoding' => 'gzip'},
-					        -proxy   => $self->proxy || ''
-					        ) or next;
-                            
+    my $fetcher = $self->make_fetcher($request) or next;
     $fetcher->debug(1) if $self->debug;
     $self->add_pending($fetcher);
   }
 
   my $timeout = $self->timeout;
 
-  # create two IO::Select objects to handle writing & reading  
+  # create two IO::Select objects to handle writing & reading
   my $readers = IO::Select->new;
   my $writers = IO::Select->new;
 
@@ -422,11 +452,16 @@ sub run_requests {
 
     foreach (@$writable) {                      # handle is ready for writing
       my $fetcher = $self->{sockets}{$_};       # recover the HTTP fetcher
-      my $result = $fetcher->send_request();      # try to send the request
-      $readers->add($_) if $result;             # send successful, so monitor for reading
-      $fetcher->request->error($fetcher->error())
-	  unless $result;                           # copy the error message
-      $writers->remove($_);                     # and remove from list monitored for writing
+      my $result = $fetcher->send_request();               # try to send the request
+      if ($result) {
+	if ($result eq 'reading header') {    # request is sent, so monitor for reading
+	  $readers->add($_);
+	  $writers->remove($_);               # and remove from list monitored for writing
+	}
+      } else {  # some sort of error
+	$fetcher->request->error($fetcher->error());       # copy the error message
+	$writers->remove($_);                              # and remove from list monitored for writing
+      }
     }
 
     foreach (@$readable) {                      # handle is ready for reading
@@ -434,8 +469,14 @@ sub run_requests {
       my $result = $fetcher->read;              # read some data
       if($fetcher->error
 	     && $fetcher->error =~ /^401\s/
-	     && $self->auth_callback()){              # Don't give up if given authentication challenge
-         $self->authenticate($fetcher);         # The result will automatically appear, as fetcher contains request reference
+	     && $self->auth_callback()) {       # Don't give up if given authentication challenge
+	# The result will automatically appear, as fetcher contains request reference
+	my $new_sock = $self->authenticate($fetcher);
+	if ($new_sock) {
+	  $writers->remove($_);
+	  $readers->remove($_);
+	  $writers->add($new_sock);
+	}
       }
       unless ($result) {                        # remove if some error occurred
 	$fetcher->request->error($fetcher->error) unless defined $result;
@@ -448,7 +489,7 @@ sub run_requests {
   # handle timeouts
   if ($timed_out) {
     while (my ($sock,$f) = each %{$self->{sockets}}) { # list of still-pending requests
-      $f->request->error('timeout');
+      $f->request->error('509 timeout');
       $readers->remove($sock);
       $writers->remove($sock);
       close $sock;
@@ -487,16 +528,20 @@ sub authenticate($$$){
 
   my ($user, $pass) = &$callback ($fetcher, $realm, ++($self->{auth_iter}->{$realm}));
 
-  if(not defined $user){  #Give up, denote with negative iteration value
+  if(!defined $user or $user eq ''){  #Give up, denote with negative iteration value
     $self->{auth_iter}->{$realm} = -1;
+    return;
   }
 
-  # Reuse request (no need to manipulate result lists) with new authentication built into dsn
+  # Reuse request, adding the authentication info
   my $request = $fetcher->request;
   $self->remove_pending($fetcher);
+
   # How do we clean up the old fetcher,which is no longer needed?
-  $fetcher->request->dsn->set_authentication($user, $pass);
-  return $self->run_requests([$request]);
+  $request->auth($user,$pass);
+  my $new_fetcher =  $self->make_fetcher($request) or return;
+  $self->add_pending($new_fetcher);
+  return $new_fetcher->socket;
 }
 
 1;
@@ -512,27 +557,48 @@ Bio::Das - Interface to Distributed Annotation System
 
   use Bio::Das;
 
+   # SERIAL API
+   my $das = Bio::Das->new(-source => 'http://www.wormbase.org/db/das',
+                           -dsn    => 'elegans',
+                           -aggregators => ['primary_transcript','clone']);
+   my $segment  = $das->segment('Chr1');
+   my @features = $segment->features;
+   my $dna      = $segment->dna;
+
   # PARALLEL API
   # create a new DAS agent with a timeout of 5 sec
   my $das = Bio::Das->new(5);
 
-  # fetch features from wormbase server spanning two segments on chromosome I
-  my $response = $das->features(-dsn     => 'http://www.wormbase.org/db/das/elegans',
-				-segment => ['CHROMOSOME_I:1,10000',
-                                             'CHROMOSOME_I:10000,20000'
-                                            ]
-			       );
-  die $response->error unless $response->is_success;
-  my $results = $response->results;
-  for my $segment (keys %$results) {
-      my @features = @{$results->{$segment}};
-      print join ' ',$seg,@features,"\n";
+  # fetch features from wormbase live and development servers spanning two segments on chromosome I
+  my @request = $das->features(-dsn     => ['http://www.wormbase.org/db/das/elegans',
+  					    'http://dev.wormbase.org/db/das/elegans',
+					   ],
+			       -segment => ['I:1,10000',
+					    'I:10000,20000'
+					   ]
+			      );
+
+  for my $request (@request) {
+    if ($request->is_success) {
+      print "\nResponse from ",$request->dsn,"\n";
+      my $results = $request->results;
+      for my $segment (keys %$results) {
+	my @features = @{$results->{$segment}};
+	print "\t",join ' ',$segment,@features,"\n";
+      }
+    }
+
+    else { #error
+      warn $request->dsn,": ",$request->error,"\n";
+    }
   }
 
-  # alternatively, invoke with a callback:
-  $das->features(-dsn     => 'http://www.wormbase.org/db/das/elegans',
-	  	 -segment => ['CHROMOSOME_I:1,10000',
-                              'CHROMOSOME_I:10000,20000'
+  # Same thing, but using a callback:
+  $das->features(-dsn     => ['http://www.wormbase.org/db/das/elegans',
+			      'http://dev.wormbase.org/db/das/elegans',
+			      ],
+	  	 -segment => ['I:1,10000',
+                              'I:10000,20000'
                              ],
 		 -callback => sub { my $feature = shift;
                                     my $segment = $feature->segment;
@@ -541,13 +607,6 @@ Bio::Das - Interface to Distributed Annotation System
                                   }
 			       );
 
-   # SERIALIZED API
-   my $das = Bio::Das->new(-source => 'http://www.wormbase.org/db/das',
-                           -dsn    => 'elegans',
-                           -aggregators => ['primary_transcript','clone']);
-   my $segment  = $das->segment('Chr1');
-   my @features = $segment->features;
-   my $dna      = $segment->dna;
 
 =head1 DESCRIPTION
 
@@ -570,10 +629,10 @@ DAS servers can be running simultaneously.
 
 =item Bio::Das::Request
 
-This class encapsulates a request to a particular DAS server, as well
-as the response that is returned from it.  Methods allow you to return
-the status of the request, the error message if any, and the data
-results.
+This class encapsulates a request to a particular DAS server.  After
+execution of the request, the response can be recovered from the
+object as well.  Methods allow you to return the status of the
+request, the error message if any, and the data results.
 
 =item Bio::Das::Segment
 
@@ -600,9 +659,44 @@ This class contains information about the stylesheet for a DAS source.
 
 =back
 
+=head2 PARALLEL AND SERIAL APIs
+
+Bio::Das supports two distinct APIs. One is a parallel API which
+allows you to make Das requests on two or more servers simultaneously.
+This is highly efficient, but the API is slightly more difficult to
+use.  The other is a serial API which supports only a single request
+on a single service.  It is recommended for simple scripts or for
+those where performance is not at a premium.
+
+The two APIs use the same objects. You select which API to use when
+you create the Das object with Bio::Das->new().
+
 =head2 OBJECT CREATION
 
-The public Bio::Das constructor is new():
+The public Bio::Das constructor is new().  It is used both for the
+parallel and serial APIs.
+
+B<Serial API object construction:>
+
+=over 4
+
+=item $das = Bio::Das->new(-server => $url, -dsn => $dsn, -aggregators=>\@aggregators);
+
+Clients that will be accessing a single server exclusively can
+indicate that they wish to use the serial APi by passing the
+B<-server> argument.  The argument for B<-server> is the base name of
+the DAS server (e.g. http://www.wormbase.org/db/das).  You may also
+select the data source to use (e.g. "elegans") by passing the B<-dsn>
+argument. B<-aggregators> is a list of aggregators as described
+earlier.
+
+=item $das = Bio::Das->new('http://das.server/cgi-bin/das',$dsn,$aggregators)
+
+Shortcut for the above.
+
+=back
+
+B<Parallel API object construction:>
 
 =over 4
 
@@ -628,17 +722,6 @@ If successful, this method returns a Bio::Das object.
 
 Shortcut for the above.
 
-=item $das = Bio::Das->new(-source => $url, -dsn => $dsn, -aggregators=>\@aggregators);
-
-This is the serialized DAS API for clients that will be accessing a
-single server exclusively.  The arguments are the URL of the remote
-DAS server (ending with the "das" component of the URL), the remote
-data source, and the list of aggregators to load. 
-
-=item $das = Bio::Das->new('http://das.server/cgi-bin/das',$dsn,$aggregators)
-
-Shortcut for the above.
-
 =back
 
 =head2 ACCESSOR METHODS
@@ -665,24 +748,175 @@ Get or set the timeout for slow servers.
 
 Get or set a flag that will turn on verbose debugging messages.
 
+=item $das->add_aggregator($aggregator)
+
+Aggregators allow you to dynamically build up more multipart features
+from the simple one-part that are returned by Das servers.  The
+concept of aggregation was introduced in the L<Bio::DB::GFF> module,
+and is completely compatible with the Bio::Das implementation.  See
+L<Bio::DB::GFF> and L<Bio::DB::GFF::Aggregator> for information on how
+to create and use aggregators.
+
+The add_aggregator() method will append an aggregator to the end of
+the list of registered aggregators.  Three different argument types
+are accepted:
+
+  1) a Bio::DB::GFF::Aggregator object -- will be added
+  2) a string in the form "aggregator_name{subpart1,subpart2,subpart3/main_method}"
+         -- will be turned into a Bio::DB::GFF::Aggregator object (the /main_method
+        part is optional).
+  3) a valid Perl token -- will be turned into a Bio::DB::GFF::Aggregator
+        subclass, where the token corresponds to the subclass name.
+
+=item $das->aggregators([@new_aggregators]);
+
+This method will get or set the list of aggregators assigned to
+the database.  If 1 or more arguments are passed, the existing
+set will be cleared.
+
+=item $das->clear_aggregators
+
+This method will clear the aggregators stored in the database object.
+Use aggregators() or add_aggregator() to add some back.
+
 =back
 
-=head2 DATA FETCHING METHODS
+=head2 DATA FETCHING METHODS - SERIAL API
 
-The following methods accept a series of arguments, contact the
-indicated DAS servers, and return a series of response objects from
-which you can learn the status of the request and fetch the results.
+We will document that serial API first, followed by the parallel API.
+Do not be confused by the fact is that both serial and parallel APIs
+have the same method names.  The behavior of the methods are
+determined solely by whether the B<-server> argument was provided to
+Bio::Das->new() during object construction.
 
 =over 4
 
-=item @response = $das->dsn(@list_of_urls)
+=item @dsn = $das->sources
+
+Return a list of data sources available from this server.  This is one
+of the few methods that can be called before setting the data source.
+
+=item $segment = $das->segment($id)
+
+=item $segment = $das->segment(-ref => $reference [,@args]);
+
+The segment() method returns a new Bio::Das::Segment object, which can
+be queried for information related to a sequence segment.  There are
+two forms of this call.  In the single-argument form, you pass
+segment() an ID to be used as the reference sequence.  Sequence IDs
+are server-specific (some servers will accept genbank accession
+numbers, others more complex IDs such as Locus:unc-9).  The method
+will return a Bio::Das::Segment object containing a region of the
+genomic corresponding to the ID.
+
+Once you fetch the segment, you can use it to fetch the features that
+overlap that segment, or the DNA corresponding to the segment.  For
+example:
+
+   my @features = $segment->features();
+   my $dna      = $segment->dna();
+
+See L<Bio::Das::Segment> for more details.
+
+Instead of a segment ID, you may use a previously-created
+Bio::Das::Segment object, in which case a copy of the segment will be
+returned to you.  You can then adjust its start and end positions.
+
+In the multiple-argument form, you pass a series of argument/value
+pairs:
+
+  Argument   Value                   Default
+  --------   -----                   -------
+
+  -ref       Reference ID            none
+  -segment   Bio::Das::Segment obj   none
+  -start     Starting position       1
+  -end       Ending position         length of ref ID
+  -offset    Starting position       0
+             (0-based)
+  -length    Length of segment       length of ref ID
+
+The B<-ref> argument is required, and indicates the ID of the genomic
+segment to retrieve.  B<-segment> is optional, and can be used to use
+a previously-created Bio::Das::Segment object as the reference point
+instead.  If both arguments are passed, B<-segment> supersedes
+B<-ref>.
+
+B<-start> and B<-end> indicate the start and stop of the desired
+genomic segment, relative to the reference ID.  If not provided, they
+default to the start and stop of the reference segment.  These
+arguments use 1-based indexing, so a B<-start> of 0 positions the
+segment one base before the start of the reference.
+
+B<-offset> and B<-length> arguments are alternative ways to indicate a
+segment using zero-based indexing.  It is probably not a good to mix
+the two calling styles, but if you do, be aware that B<-offset>
+supersedes B<-start> and B<-length> supersedes B<-stop>.
+
+Note that no checking of the validity of the passed reference ID will
+be performed until you call the segment's features() or dna() methods.
+
+=item @segments = $das->get_feature_by_name(-name=>$name [,-class=>$class]);
+
+This method implements the DAS feature request using parameters that
+will translate a feature name into one or more segments.  This can be
+used to retrieve the section of a genome that is occupied by a
+particular feature.  If the feature name matches multiple features in
+discontinuous parts of the genome, this call may return multiple
+segments.  Once you have a segment, you can call its features() method
+to get information about the features that overlap this region.
+
+The optional -class argument is provided to deal with servers that
+have namespaced their features using a colon.
+$das->get_feature_by_name(-name=>'foo',-class=>'bar') is exactly
+equivalent to $das->get_feature_by_name(-name=>'bar:foo').
+
+Because this method is misnamed (it returns segments, not features),
+it is also known as feature2segment().
+
+The method can also be called using the shortcut syntax
+get_feature_by_name($name).
+
+=item @entry_points = $das->entry_points
+
+The entry_points() method returns an array of Bio::Das::Segment
+objects that have been designated "entry points" by the DAS server.
+Also see the Bio::Das::Segment->entry_points() method.
+
+=item $stylesheet = $das->stylesheet
+
+Return the stylesheet from the remote DAS server.  The stylesheet
+contains suggestions for the visual format for the various features
+provided by the server and can be used to translate features into
+glyphs.  The object returned is a Bio::Das::Stylesheet object.
+
+=item @types = $das->types
+
+This method returns a list of all the annotation feature types served
+by the DAS server.  The return value is an array of Bio::Das::Type
+objects.
+
+=back
+
+=head2 DATA FETCHING METHODS - PARALLEL API
+
+The following methods accept a series of arguments, contact the
+indicated DAS servers, and return a series of request objects from
+which you can learn the status of the request and fetch the results.
+
+Parallel API:
+
+=over 4
+
+=item @request = $das->dsn(@list_of_urls)
 
 The dsn() method accepts a list of DAS server URLs and returns a list
-of the DSNs provided by each server.
+of request objects containing the DSNs provided by each server.
 
 The request objects will indicate whether each request was successful
 via their is_success() methods.  For your convenience, the request
-object is automagically stringified into the requested URL.  For example:
+object is automagically stringified into the requested URL.  For
+example:
 
  my $das = Bio::Das->new(5);  # timeout of 5 sec
  my @response = $das->dsn('http://stein.cshl.org/perl/das',
@@ -738,7 +972,7 @@ Notice that the DSN URLs always have the format:
 
 In which the ^^^ indicated part is identical to the server address.
 
-=item @response = $das->types(-dsn=>[$dsn1,$dsn2],@other_args)
+=item @request = $das->types(-dsn=>[$dsn1,$dsn2],@other_args)
 
 The types() method asks the indicated servers to return the feature
 types that they provide.  Arguments are name-value pairs:
@@ -759,13 +993,14 @@ types that they provide.  Arguments are name-value pairs:
                 categories.
 
   -enumerate    (optional) If true, the server will return the count of
-                each time.
+                each type.  The count can be retrieved using the 
+                L<Bio::Das::Type> objects' count() method.
 
   -callback     (optional) Specifies a subroutine to be invoked on each
                 type object received.
 
 Segments have the format: "seq_id:start,end".  If successful, the
-response results() method will return a list of Bio::Das::Type
+request results() method will return a list of L<Bio::Das::Type>
 objects.
 
 If a callback is specified, the code ref will be invoked with two
@@ -774,10 +1009,10 @@ the second is an array ref containing the list of types present in
 that segment.  If no -segment argument was provided, then the callback
 will be invoked once with a dummy segment (a version, but no seq_id,
 start or end), and an arrayref containing the types.  If a callback is
-specified, then the @response array will return the status codes for
+specified, then the @request array will return the status codes for
 each request, but invoking results() will return empty.
 
-=item @response = $das->entry_points(-dsn=>[$dsn1,$dsn2],@other_args)
+=item @request = $das->entry_points(-dsn=>[$dsn1,$dsn2],@other_args)
 
 Invoke an entry_points request.  Arguments are name-value pairs:
 
@@ -792,17 +1027,17 @@ Invoke an entry_points request.  Arguments are name-value pairs:
   -callback     (optional) Specifies a subroutine to be invoked on each
                 segment object received.
 
-If a callback is specified, then the @response array will contain the
+If a callback is specified, then the @request array will contain the
 status codes for each request, but the results() method will return
 empty.
 
-Successful responses will return a set of Bio::Das::Segment objects.
+Successful requests will return a set of Bio::Das::Segment objects.
 
-=item @response = $das->features(-dsn=>[$dsn1,$dsn2],@other_args)
+=item @request = $das->features(-dsn=>[$dsn1,$dsn2],@other_args)
 
 Invoke a features request to return a set of Bio::Das::Feature
-objects.  The -dsn argument is required, and may point to a single DSN
-or to an array ref of several DSNs.  Other arguments are optional:
+objects.  The B<-dsn> argument is required, and may point to a single
+DSN or to an array ref of several DSNs.  Other arguments are optional:
 
   Argument      Description
   --------      -----------
@@ -849,8 +1084,16 @@ or to an array ref of several DSNs.  Other arguments are optional:
   -iterator     (optional)  If true, specifies that an iterator should be
                 returned rather than a list of features.
 
+The features() method returns a list of L<Bio::Das::Request> objects.
+There will be one request for each DAS DSN provided in the B<-dsn>
+argument.  Requests are returned in the same order that they were
+passed to B<-dsn>, but you can also query the Bio::Das::Request
+object to determine which server processed the request.  See Fetching
+Results for details.  If you happen to call this method in a scalar
+context, it will return the first request, discarding the rest.
+
 If a callback (-callback or -segment_callback) is specified, then the
-@response array will contain the status codes for each request, but
+@request array will contain the status codes for each request, but
 results() will return empty.
 
 The subroutine specified by -callback will be invoked every time a
@@ -880,10 +1123,12 @@ a single method, next_seq(), which returns the next Feature.  Example:
 
    $iterator = $das->features(-dsn=>[$dsn1,$dsn2],-iterator=>1);
    while (my $feature = $iterator->next_seq) {
-     print "got a ",$feature->method,"\n";
+     my $dsn  = $feature->segment->dsn;
+     my $type = $feature->type;
+     print "got a $type from $dsn\n";
    }
 
-=item @response = $das->dna(-dsn=>[$dsn1,$dsn2],@other_args)
+=item @request = $das->dna(-dsn=>[$dsn1,$dsn2],@other_args)
 
 Invoke a features request to return a DNA string.  The -dsn argument
 is required, and may point to a single DSN or to an array ref of
@@ -907,68 +1152,239 @@ several DSNs.  Other arguments are optional:
 -dsn, -segment and -callback have the same meaning that they do in
 similar methods.
 
+=item @request = $das->stylesheet(-dsn=>[$dsn1,$dsn2],@other_args)
+
+Invoke a stylesheet request to return the L<Bio::Das::Stylesheet>
+object.  The -dsn argument is required, and may point to a single DSN
+or to an array ref of several DSNs.  Other arguments are optional:
+
+  Argument      Description
+  --------      -----------
+
+  -dsn          A DAS DSN, as returned by the dsn() call.  You may
+                also provide a simple string containing the DSN URL.
+                To make the types() request on multiple servers, pass an
+                array reference containing the list of DSNs.
+
+  -segment      (optional) A single segment, or an array ref containing
+                several segments.  Segments are either Bio::Das::Segment
+                objects, or strings of the form "seq_id:start,end".
+
+  -callback     (optional) Specifies a subroutine to be invoked on each
+               stylesheet received.
+
+-dsn, -segment and -callback have the same meaning that they do in
+similar methods.
+
+=item @request = $das->get_feature_by_name(-dsn=>[$dsns],-name=>$name [,-class=>$class]);
+
+This method implements the DAS feature request using parameters that
+will translate a feature name into one or more segments.  This can be
+used to retrieve the section of a genome that is occupied by a
+particular feature.  If the feature name matches multiple features in
+discontinuous parts of the genome, this call may return multiple
+segments.  Once you have a segment, you can call its features() method
+to get information about the features that overlap this region.
+
+The optional -class argument is provided to deal with servers that
+have namespaced their features using a colon.
+$das->get_feature_by_name(-name=>'foo',-class=>'bar') is exactly
+equivalent to $das->get_feature_by_name(-name=>'bar:foo').
+
+Because this method is misnamed (it returns segments, not features),
+it is also known as feature2segment().
+
+In case of a successful request, the request results() method will
+return a list of Bio::Das::Segment objects, which can then be passed
+back to features().
+
 =back
 
-=head2 add_aggregator
+=head2 Fetching Results
 
-NOTE: Aggregator support is currently experimental and is provided for
-compatibility with Generic Genome Browser.
+When using the parallel API, the dsn(), features(), dna(), and
+stylesheet() methods will return an array of L<Bio::Das::Request>
+objects.  Each object contains information about the outcome of the
+request and the results, if any, returned.  The request objects
+correspond to each of the DSNs passed to the request in the B<-dsn>
+argument, and have the same number and order.
 
- Title   : add_aggregator
- Usage   : $db->add_aggregator($aggregator)
- Function: add an aggregator to the list
- Returns : nothing
- Args    : an aggregator
- Status  : public
+Because of the inherent uncertainties of the Internet, any DAS request
+can fail.  It could fail because of a network transmission error, a
+timeout, a down server, an HTTP URL-not-found error, or an unparseable
+DAS document.  For this reason, you should check each request's
+is_success() method before trying to use the results.  Here is the
+canonical code:
 
-This method will append an aggregator to the end of the list of
-registered aggregators.  Three different argument types are accepted:
+  my @requests = $das->some_method(-dsn=>[$dsn1,$dsn2,$dsn3]);
+  for my $request (@requests) {
+    if ($request->is_success) {
+       my $results = $request->results;
+       # do something with the results
+    }
 
-  1) a Bio::DB::GFF::Aggregator object -- will be added
-  2) a string in the form "aggregator_name{subpart1,subpart2,subpart3/main_method}"
-         -- will be turned into a Bio::DB::GFF::Aggregator object (the /main_method
-        part is optional).
-  3) a valid Perl token -- will be turned into a Bio::DB::GFF::Aggregator
-        subclass, where the token corresponds to the subclass name.
+    else {
+       warn $request->error;
+    }
+  }
 
-=cut
+The is_success() method returns true on a successful request, false
+otherwise.  In case of an unsuccessful request, the error() method
+will provide additional information on why the request failed The
+format is "XXXX human-readable string" as in:
 
-=head2 aggregators
+    400 Bad command
 
- Title   : aggregators
- Usage   : $db->aggregators([@new_aggregators]);
- Function: retrieve list of aggregators
- Returns : list of aggregators
- Args    : a list of aggregators to set (optional)
- Status  : public
+The following error strings can be returned:
 
-This method will get or set the list of aggregators assigned to
-the database.  If 1 or more arguments are passed, the existing
-set will be cleared.
+       400 Bad command
+       401 Bad data source
+       402 Bad command arguments
+       403 Bad reference object
+       404 Bad stylesheet
+       405 Coordinate error
+       410 Unknown host
+       411 Couldn't connect
+       412 Communications error
+       413 Authentication scheme 'xxxx" is not supported
+       500 Server error
+       501 Unimplemented feature
+       502 No X-Das-Version header
+       503 Invalid X-Das-Version header
+       504 DAS server is too old
+       505 No X-Das-Status header
+       506 Data decompression failure
 
-=cut
+To discover which server a request was sent to, you can call its dsn()
+method.  This will return the server and data source as a single URL,
+e.g.:
 
-=head2 clear_aggregators
+   my $dsn = $request->dsn;
+   print $dsn,"\n";  # prints 'http://www.wormbase.org/db/das/elegans'
 
- Title   : clear_aggregators
- Usage   : $db->clear_aggregators
- Function: clears list of aggregators
- Returns : nothing
- Args    : none
- Status  : public
+What is returned is actually a L<Bio::Das::DSN> object.  You can call
+the object's base() method to return the server part of the DSN, and
+its id() method to return the data source:
 
-This method will clear the aggregators stored in the database object.
-Use aggregators() or add_aggregator() to add some back.
+   my $dsn = $request->dsn;
+   print $dsn->base,"\n";  # prints 'http://www.wormbase.org/db/das'
+   print $dsn->id,"\n";    # prints 'elegans'
 
-=cut
+To get the results of from the request, call its results() method.  In
+a list context, results() will return a list of the appropriate
+objects for the request (a set of L<Bio::Das::Feature> objects for the
+features() request a set of L<Bio::Das::Stylesheet> objects for the
+stylesheet() request, a set of L<Bio::Das::Type> objects for the
+types() request, and a set of raw DNA strings for the dna()
+request.)
 
-=head2 Fetching results
+In a scalar context, results() will return a hashref in which the keys
+are the segment strings passed to the request with the B<-segments>
+argument and the values are arrayrefs containing the list of results.
 
-- documentation pending -
+There is an equivalence here.  When this code fragment executes, both
+$results_hash1 and $results_hash2 will contain the same information.
+
+  my @results = $request->results;
+  my $result_hash1 = {};
+  for my $r (@results) {
+     my $segment = $r->segment;
+     push @{$result_hash{$segment}},$r;
+  }
+
+  my $result2_hash2 = $request->results;
 
 =head2 Authentication
 
-- documentation pending -
+It may be desirable to access DAS data that is stored in an
+authenticating (password protected) server.  Only HTTP Basic
+authentication is currently supported by Bio::Das, but you can run the
+authentication over an SSL connection, thereby avoiding the risk of
+passwords being sniffed.
+
+Authentication information can be passed to the server in either of
+two ways:
+
+=over 4
+
+=item In the server's URL
+
+You can provide the username and password in the form:
+
+   http://user:pass@my.das.server.org/cgi-bin/das
+
+Where B<user> and B<pass> are the username and password required for
+authentication.
+
+Unless you do with this an SSL (https:) connection, you will get a
+warning that using the password in the URL violates the recommendation
+in RFC 2396.  You can suppress this warning using the no_rfc_warning()
+method:
+
+  $das->no_rfc_warning(1);
+
+=item Using an authentication callback
+
+You can provide a subroutine code reference that returns the username
+and password at the time you create the Bio::Das object.  When
+accessing a password protected site, Bio::Das will invoke your
+callback using information about the request. The callback will return
+the appropriate username and password.  You can do whatever you need
+to do to get the authentication information, whether accessing an
+enterprise database, or popping up a dialog box for the user to
+respond to.
+
+=back
+
+To install an authentication callback, pass a coderef to
+the B<-auth_callback> argument when calling Bio::Das->new():
+
+  Bio::Das->new(-auth_callback=>\&my_authentication_routine);
+
+The callback will be called with three arguments:
+
+  my_authentication_routine($fetcher,$realm,$iteration_count)
+
+B<$fetcher> is an L<Bio::Das::HTTP::Fetch> object.  It contains the
+information you will need to determine which server is requesting
+authentication.  You will probably want to call the fetch object's
+host() method to get the name of the DAS host, but if you require more
+information, the request() method will return the L<Bio::Das::Request>
+object with complete information about the request.
+
+B<$realm> is the Basic Authentication Realm string, as returned by the
+remote server.
+
+B<$iteration_count> records the number of times your authentication
+routine has been invoked for this particular realm.  You can use this
+information to abort authentication if it fails the first time.
+
+The authentication callback should a two-element list containing the
+username and password for authentication against the server.  If it
+returns an empty list, the request will be aborted.
+
+Here is a sample authentication routine.  It prompts the user up to
+three times for his username and password, and then aborts.  Notice
+the way in which the hostname is recovered from the
+Bio::Das::HTTP::Fetch object.
+
+ sub my_authentication_routine {
+   my ($fetcher,$domain,$iteration_count) = @_;
+   return if $iteration_count > 3;
+   my $host = $fetcher->request->host;
+   print STDERR "$host/$domain requires authentication (try $iteration_count of 3)\n";
+   print STDERR "Username: ";
+   chomp (my $username = <>);
+   print STDERR "Password: ";
+   chomp (my $password = <>);
+   return ($username,$password);
+ }
+
+Note: while processing the authentication callback, processing of
+other pending requests will stall, usually at the point at which the
+request has been sent, but the results have not yet been received and
+parsed.  For this reason, you might want to include a timeout in your
+authentication routine.
 
 =head1 AUTHOR
 
@@ -982,6 +1398,7 @@ disclaimers of warranty.
 
 =head1 SEE ALSO
 
+L<Bio::Das::Request>, L<Bio::Das::HTTP::Fetch>,
 L<Bio::Das::Segment>, L<Bio::Das::Type>, L<Bio::Das::Stylesheet>,
 L<Bio::Das::Source>, L<Bio::RangeI>
 
